# Copyright (C) 2008-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Fetching from git into bzr."""

import posixpath
import stat

from dulwich.object_store import tree_lookup_path
from dulwich.objects import S_IFGITLINK, S_ISGITLINK, ZERO_SHA, Commit, Tag, Tree

from .. import debug, osutils, trace
from ..bzr.inventory import (
    InventoryDirectory,
    InventoryFile,
    InventoryLink,
    TreeReference,
)
from ..bzr.inventorytree import InventoryRevisionTree
from ..bzr.testament import StrictTestament3
from ..bzr.versionedfile import ChunkedContentFactory
from ..errors import BzrError
from ..revision import NULL_REVISION
from ..transport import NoSuchFile
from ..tree import InterTree
from ..tsort import topo_sort
from .mapping import (
    DEFAULT_FILE_MODE,
    decode_git_path,
    mode_is_executable,
    mode_kind,
    warn_unusual_mode,
)
from .object_store import LRUTreeCache, _tree_to_objects


def import_git_blob(
    texts,
    mapping,
    path,
    name,
    hexshas,
    base_bzr_tree,
    parent_id,
    revision_id,
    parent_bzr_trees,
    lookup_object,
    modes,
    store_updater,
    lookup_file_id,
):
    """Import a git blob object into a bzr repository.

    :param texts: VersionedFiles to add to
    :param path: Path in the tree
    :param blob: A git blob
    :return: Inventory delta for this file
    """
    if not isinstance(path, bytes):
        raise TypeError(path)
    decoded_path = decode_git_path(path)
    (base_mode, mode) = modes
    (base_hexsha, hexsha) = hexshas
    if mapping.is_special_file(path):
        return []
    if base_hexsha == hexsha and base_mode == mode:
        # If nothing has changed since the base revision, we're done
        return []
    file_id = lookup_file_id(decoded_path)
    if stat.S_ISLNK(mode):
        cls = InventoryLink
    else:
        cls = InventoryFile
    ie = cls(file_id, decode_git_path(name), parent_id)
    if ie.kind == "file":
        ie.executable = mode_is_executable(mode)
    if base_hexsha == hexsha and mode_kind(base_mode) == mode_kind(mode):
        base_exec = base_bzr_tree.is_executable(decoded_path)
        if ie.kind == "symlink":
            ie.symlink_target = base_bzr_tree.get_symlink_target(decoded_path)
        else:
            ie.text_size = base_bzr_tree.get_file_size(decoded_path)
            ie.text_sha1 = base_bzr_tree.get_file_sha1(decoded_path)
        if ie.kind == "symlink" or ie.executable == base_exec:
            ie.revision = base_bzr_tree.get_file_revision(decoded_path)
        else:
            blob = lookup_object(hexsha)
    else:
        blob = lookup_object(hexsha)
        if ie.kind == "symlink":
            ie.revision = None
            ie.symlink_target = decode_git_path(blob.data)
        else:
            ie.text_size = sum(map(len, blob.chunked))
            ie.text_sha1 = osutils.sha_strings(blob.chunked)
    # Check what revision we should store
    parent_keys = []
    for ptree in parent_bzr_trees:
        intertree = InterTree.get(ptree, base_bzr_tree)
        try:
            ppath = intertree.find_source_paths(decoded_path, recurse="none")
        except NoSuchFile:
            continue
        if ppath is None:
            continue
        pkind = ptree.kind(ppath)
        if pkind == ie.kind and (
            (
                pkind == "symlink"
                and ptree.get_symlink_target(ppath) == ie.symlink_target
            )
            or (
                pkind == "file"
                and ptree.get_file_sha1(ppath) == ie.text_sha1
                and ptree.is_executable(ppath) == ie.executable
            )
        ):
            # found a revision in one of the parents to use
            ie.revision = ptree.get_file_revision(ppath)
            break
        parent_key = (file_id, ptree.get_file_revision(ppath))
        if parent_key not in parent_keys:
            parent_keys.append(parent_key)
    if ie.revision is None:
        # Need to store a new revision
        ie.revision = revision_id
        if ie.revision is None:
            raise ValueError("no file revision set")
        if ie.kind == "symlink":
            chunks = []
        else:
            chunks = blob.chunked
        texts.insert_record_stream(
            [
                ChunkedContentFactory(
                    (file_id, ie.revision), tuple(parent_keys), ie.text_sha1, chunks
                )
            ]
        )
    invdelta = []
    if base_hexsha is not None:
        old_path = decoded_path  # Renames are not supported yet
        if stat.S_ISDIR(base_mode):
            invdelta.extend(
                remove_disappeared_children(
                    base_bzr_tree,
                    old_path,
                    lookup_object(base_hexsha),
                    [],
                    lookup_object,
                )
            )
    else:
        old_path = None
    invdelta.append((old_path, decoded_path, file_id, ie))
    if base_hexsha != hexsha:
        store_updater.add_object(blob, (ie.file_id, ie.revision), path)
    return invdelta


class SubmodulesRequireSubtrees(BzrError):
    _fmt = (
        "The repository you are fetching from contains submodules, "
        "which require a Bazaar format that supports tree references."
    )
    internal = False


def import_git_submodule(
    texts,
    mapping,
    path,
    name,
    hexshas,
    base_bzr_tree,
    parent_id,
    revision_id,
    parent_bzr_trees,
    lookup_object,
    modes,
    store_updater,
    lookup_file_id,
):
    """Import a git submodule."""
    (base_hexsha, hexsha) = hexshas
    (base_mode, mode) = modes
    if base_hexsha == hexsha and base_mode == mode:
        return [], {}
    path = decode_git_path(path)
    file_id = lookup_file_id(path)
    invdelta = []
    ie = TreeReference(file_id, decode_git_path(name), parent_id)
    ie.revision = revision_id
    if base_hexsha is not None:
        old_path = path  # Renames are not supported yet
        if stat.S_ISDIR(base_mode):
            invdelta.extend(
                remove_disappeared_children(
                    base_bzr_tree,
                    old_path,
                    lookup_object(base_hexsha),
                    [],
                    lookup_object,
                )
            )
    else:
        old_path = None
    ie.reference_revision = mapping.revision_id_foreign_to_bzr(hexsha)
    texts.insert_record_stream(
        [ChunkedContentFactory((file_id, ie.revision), (), None, [])]
    )
    invdelta.append((old_path, path, file_id, ie))
    return invdelta, {}


def remove_disappeared_children(
    base_bzr_tree, path, base_tree, existing_children, lookup_object
):
    """Generate an inventory delta for removed children.

    :param base_bzr_tree: Base bzr tree against which to generate the
        inventory delta.
    :param path: Path to process (unicode)
    :param base_tree: Git Tree base object
    :param existing_children: Children that still exist
    :param lookup_object: Lookup a git object by its SHA1
    :return: Inventory delta, as list
    """
    if not isinstance(path, str):
        raise TypeError(path)
    ret = []
    for name, mode, hexsha in base_tree.iteritems():
        if name in existing_children:
            continue
        c_path = posixpath.join(path, decode_git_path(name))
        file_id = base_bzr_tree.path2id(c_path)
        if file_id is None:
            raise TypeError(file_id)
        ret.append((c_path, None, file_id, None))
        if stat.S_ISDIR(mode):
            ret.extend(
                remove_disappeared_children(
                    base_bzr_tree, c_path, lookup_object(hexsha), [], lookup_object
                )
            )
    return ret


def import_git_tree(
    texts,
    mapping,
    path,
    name,
    hexshas,
    base_bzr_tree,
    parent_id,
    revision_id,
    parent_bzr_trees,
    lookup_object,
    modes,
    store_updater,
    lookup_file_id,
    allow_submodules=False,
):
    """Import a git tree object into a bzr repository.

    :param texts: VersionedFiles object to add to
    :param path: Path in the tree (str)
    :param name: Name of the tree (str)
    :param tree: A git tree object
    :param base_bzr_tree: Base inventory against which to return inventory
        delta
    :return: Inventory delta for this subtree
    """
    (base_hexsha, hexsha) = hexshas
    (base_mode, mode) = modes
    if not isinstance(path, bytes):
        raise TypeError(path)
    if not isinstance(name, bytes):
        raise TypeError(name)
    if base_hexsha == hexsha and base_mode == mode:
        # If nothing has changed since the base revision, we're done
        return [], {}
    invdelta = []
    file_id = lookup_file_id(osutils.safe_unicode(path))
    ie = InventoryDirectory(file_id, decode_git_path(name), parent_id)
    tree = lookup_object(hexsha)
    if base_hexsha is None:
        base_tree = None
        old_path = None  # Newly appeared here
    else:
        base_tree = lookup_object(base_hexsha)
        old_path = decode_git_path(path)  # Renames aren't supported yet
    new_path = decode_git_path(path)
    if base_tree is None or type(base_tree) is not Tree:
        ie.revision = revision_id
        invdelta.append((old_path, new_path, ie.file_id, ie))
        texts.insert_record_stream(
            [ChunkedContentFactory((ie.file_id, ie.revision), (), None, [])]
        )
    # Remember for next time
    existing_children = set()
    child_modes = {}
    for name, child_mode, child_hexsha in tree.iteritems():
        existing_children.add(name)
        child_path = posixpath.join(path, name)
        if type(base_tree) is Tree:
            try:
                child_base_mode, child_base_hexsha = base_tree[name]
            except KeyError:
                child_base_hexsha = None
                child_base_mode = 0
        else:
            child_base_hexsha = None
            child_base_mode = 0
        if stat.S_ISDIR(child_mode):
            subinvdelta, grandchildmodes = import_git_tree(
                texts,
                mapping,
                child_path,
                name,
                (child_base_hexsha, child_hexsha),
                base_bzr_tree,
                file_id,
                revision_id,
                parent_bzr_trees,
                lookup_object,
                (child_base_mode, child_mode),
                store_updater,
                lookup_file_id,
                allow_submodules=allow_submodules,
            )
        elif S_ISGITLINK(child_mode):  # submodule
            if not allow_submodules:
                raise SubmodulesRequireSubtrees()
            subinvdelta, grandchildmodes = import_git_submodule(
                texts,
                mapping,
                child_path,
                name,
                (child_base_hexsha, child_hexsha),
                base_bzr_tree,
                file_id,
                revision_id,
                parent_bzr_trees,
                lookup_object,
                (child_base_mode, child_mode),
                store_updater,
                lookup_file_id,
            )
        else:
            if not mapping.is_special_file(name):
                subinvdelta = import_git_blob(
                    texts,
                    mapping,
                    child_path,
                    name,
                    (child_base_hexsha, child_hexsha),
                    base_bzr_tree,
                    file_id,
                    revision_id,
                    parent_bzr_trees,
                    lookup_object,
                    (child_base_mode, child_mode),
                    store_updater,
                    lookup_file_id,
                )
            else:
                subinvdelta = []
            grandchildmodes = {}
        child_modes.update(grandchildmodes)
        invdelta.extend(subinvdelta)
        if child_mode not in (
            stat.S_IFDIR,
            DEFAULT_FILE_MODE,
            stat.S_IFLNK,
            DEFAULT_FILE_MODE | 0o111,
            S_IFGITLINK,
        ):
            child_modes[child_path] = child_mode
    # Remove any children that have disappeared
    if base_tree is not None and type(base_tree) is Tree:
        invdelta.extend(
            remove_disappeared_children(
                base_bzr_tree, old_path, base_tree, existing_children, lookup_object
            )
        )
    store_updater.add_object(tree, (file_id, revision_id), path)
    return invdelta, child_modes


def verify_commit_reconstruction(
    target_git_object_retriever,
    lookup_object,
    o,
    rev,
    ret_tree,
    parent_trees,
    mapping,
    unusual_modes,
    verifiers,
):
    new_unusual_modes = mapping.export_unusual_file_modes(rev)
    if new_unusual_modes != unusual_modes:
        raise AssertionError(
            "unusual modes don't match: {!r} != {!r}".format(
                unusual_modes, new_unusual_modes
            )
        )
    # Verify that we can reconstruct the commit properly
    rec_o = target_git_object_retriever._reconstruct_commit(
        rev, o.tree, True, verifiers
    )
    if rec_o != o:
        raise AssertionError(
            "Reconstructed commit differs: {!r} != {!r}".format(rec_o, o)
        )
    diff = []
    new_objs = {}
    for path, obj, _ie in _tree_to_objects(
        ret_tree,
        parent_trees,
        target_git_object_retriever._cache.idmap,
        unusual_modes,
        mapping.BZR_DUMMY_FILE,
    ):
        old_obj_id = tree_lookup_path(lookup_object, o.tree, path)[1]
        new_objs[path] = obj
        if obj.id != old_obj_id:
            diff.append((path, lookup_object(old_obj_id), obj))
    for path, old_obj, new_obj in diff:
        while (
            old_obj.type_name == "tree"
            and new_obj.type_name == "tree"
            and sorted(old_obj) == sorted(new_obj)
        ):
            for name in old_obj:
                if old_obj[name][0] != new_obj[name][0]:
                    raise AssertionError(
                        "Modes for {} differ: {:o} != {:o}".format(
                            path, old_obj[name][0], new_obj[name][0]
                        )
                    )
                if old_obj[name][1] != new_obj[name][1]:
                    # Found a differing child, delve deeper
                    path = posixpath.join(path, name)
                    old_obj = lookup_object(old_obj[name][1])
                    new_obj = new_objs[path]
                    break
        raise AssertionError(
            "objects differ for {}: {!r} != {!r}".format(path, old_obj, new_obj)
        )


def ensure_inventories_in_repo(repo, trees):
    real_inv_vf = repo.inventories.without_fallbacks()
    for t in trees:
        revid = t.get_revision_id()
        if not real_inv_vf.get_parent_map([(revid,)]):
            repo.add_inventory(revid, t.root_inventory, t.get_parent_ids())


def import_git_commit(
    repo, mapping, head, lookup_object, target_git_object_retriever, trees_cache, strict
):
    o = lookup_object(head)
    # Note that this uses mapping.revision_id_foreign_to_bzr. If the parents
    # were bzr roundtripped revisions they would be specified in the
    # roundtrip data.
    rev, roundtrip_revid, verifiers = mapping.import_commit(
        o, mapping.revision_id_foreign_to_bzr, strict
    )
    if roundtrip_revid is not None:
        original_revid = rev.revision_id
        rev.revision_id = roundtrip_revid
    # We have to do this here, since we have to walk the tree and
    # we need to make sure to import the blobs / trees with the right
    # path; this may involve adding them more than once.
    parent_trees = trees_cache.revision_trees(rev.parent_ids)
    ensure_inventories_in_repo(repo, parent_trees)
    if parent_trees == []:
        base_bzr_tree = trees_cache.revision_tree(NULL_REVISION)
        base_tree = None
        base_mode = None
    else:
        base_bzr_tree = parent_trees[0]
        base_tree = lookup_object(o.parents[0]).tree
        base_mode = stat.S_IFDIR
    store_updater = target_git_object_retriever._get_updater(rev)
    inv_delta, unusual_modes = import_git_tree(
        repo.texts,
        mapping,
        b"",
        b"",
        (base_tree, o.tree),
        base_bzr_tree,
        None,
        rev.revision_id,
        parent_trees,
        lookup_object,
        (base_mode, stat.S_IFDIR),
        store_updater,
        mapping.generate_file_id,
        allow_submodules=repo._format.supports_tree_reference,
    )
    if unusual_modes != {}:
        for path, mode in unusual_modes.iteritems():
            warn_unusual_mode(rev.foreign_revid, path, mode)
        mapping.import_unusual_file_modes(rev, unusual_modes)
    try:
        basis_id = rev.parent_ids[0]
    except IndexError:
        basis_id = NULL_REVISION
        base_bzr_inventory = None
    else:
        base_bzr_inventory = base_bzr_tree.root_inventory
    rev.inventory_sha1, inv = repo.add_inventory_by_delta(
        basis_id, inv_delta, rev.revision_id, rev.parent_ids, base_bzr_inventory
    )
    ret_tree = InventoryRevisionTree(repo, inv, rev.revision_id)
    # Check verifiers
    if verifiers and roundtrip_revid is not None:
        testament = StrictTestament3(rev, ret_tree)
        calculated_verifiers = {"testament3-sha1": testament.as_sha1()}
        if calculated_verifiers != verifiers:
            trace.mutter(
                "Testament SHA1 %r for %r did not match %r.",
                calculated_verifiers["testament3-sha1"],
                rev.revision_id,
                verifiers["testament3-sha1"],
            )
            rev.revision_id = original_revid
            rev.inventory_sha1, inv = repo.add_inventory_by_delta(
                basis_id, inv_delta, rev.revision_id, rev.parent_ids, base_bzr_tree
            )
            ret_tree = InventoryRevisionTree(repo, inv, rev.revision_id)
    else:
        calculated_verifiers = {}
    store_updater.add_object(o, calculated_verifiers, None)
    store_updater.finish()
    trees_cache.add(ret_tree)
    repo.add_revision(rev.revision_id, rev)
    if "verify" in debug.debug_flags:
        verify_commit_reconstruction(
            target_git_object_retriever,
            lookup_object,
            o,
            rev,
            ret_tree,
            parent_trees,
            mapping,
            unusual_modes,
            verifiers,
        )


def import_git_objects(
    repo, mapping, object_iter, target_git_object_retriever, heads, pb=None, limit=None
):
    """Import a set of git objects into a bzr repository.

    :param repo: Target Bazaar repository
    :param mapping: Mapping to use
    :param object_iter: Iterator over Git objects.
    :return: Tuple with pack hints and last imported revision id
    """

    def lookup_object(sha):
        try:
            return object_iter[sha]
        except KeyError:
            return target_git_object_retriever[sha]

    graph = []
    checked = set()
    heads = list(set(heads))
    trees_cache = LRUTreeCache(repo)
    # Find and convert commit objects
    while heads:
        if pb is not None:
            pb.update("finding revisions to fetch", len(graph), None)
        head = heads.pop()
        if head == ZERO_SHA:
            continue
        if not isinstance(head, bytes):
            raise TypeError(head)
        try:
            o = lookup_object(head)
        except KeyError:
            continue
        if isinstance(o, Commit):
            rev, roundtrip_revid, verifiers = mapping.import_commit(
                o, mapping.revision_id_foreign_to_bzr, strict=True
            )
            if repo.has_revision(rev.revision_id) or (
                roundtrip_revid and repo.has_revision(roundtrip_revid)
            ):
                continue
            graph.append((o.id, o.parents))
            heads.extend([p for p in o.parents if p not in checked])
        elif isinstance(o, Tag):
            if o.object[1] not in checked:
                heads.append(o.object[1])
        else:
            trace.warning("Unable to import head object {!r}".format(o))
        checked.add(o.id)
    del checked
    # Order the revisions
    # Create the inventory objects
    batch_size = 1000
    revision_ids = topo_sort(graph)
    pack_hints = []
    if limit is not None:
        revision_ids = revision_ids[:limit]
    last_imported = None
    for offset in range(0, len(revision_ids), batch_size):
        target_git_object_retriever.start_write_group()
        try:
            repo.start_write_group()
            try:
                for i, head in enumerate(revision_ids[offset : offset + batch_size]):
                    if pb is not None:
                        pb.update("fetching revisions", offset + i, len(revision_ids))
                    import_git_commit(
                        repo,
                        mapping,
                        head,
                        lookup_object,
                        target_git_object_retriever,
                        trees_cache,
                        strict=True,
                    )
                    last_imported = head
            except BaseException:
                repo.abort_write_group()
                raise
            else:
                hint = repo.commit_write_group()
                if hint is not None:
                    pack_hints.extend(hint)
        except BaseException:
            target_git_object_retriever.abort_write_group()
            raise
        else:
            target_git_object_retriever.commit_write_group()
    return pack_hints, last_imported


class DetermineWantsRecorder:
    def __init__(self, actual):
        self.actual = actual
        self.wants = []
        self.remote_refs = {}

    def __call__(self, refs):
        if type(refs) is not dict:
            raise TypeError(refs)
        self.remote_refs = refs
        self.wants = self.actual(refs)
        return self.wants



# Copyright (C) 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test symlink support."""

import os

from breezy import osutils, workingtree
from breezy.tests import features
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestSmartAddTree(TestCaseWithWorkingTree):
    # See eg <https://bugs.launchpad.net/bzr/+bug/192859>

    def setUp(self):
        super().setUp()
        self.requireFeature(features.SymlinkFeature(self.test_dir))

    def test_smart_add_symlink(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents(
            [
                ("tree/link@", b"target"),
            ]
        )
        tree.smart_add(["tree/link"])
        self.assertTrue(tree.is_versioned("link"))
        self.assertFalse(tree.is_versioned("target"))
        self.assertEqual("symlink", tree.kind("link"))

    def test_smart_add_symlink_pointing_outside(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents(
            [
                ("tree/link@", "../../../../target"),
            ]
        )
        tree.smart_add(["tree/link"])
        self.assertTrue(tree.is_versioned("link"))
        self.assertFalse(tree.is_versioned("target"))
        self.assertEqual("symlink", tree.kind("link"))

    def test_add_file_under_symlink(self):
        # similar to
        # https://bugs.launchpad.net/bzr/+bug/192859/comments/3
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents(
            [
                ("tree/link@", "dir"),
                ("tree/dir/",),
                ("tree/dir/file", b"content"),
            ]
        )
        if tree.has_versioned_directories():
            self.assertEqual(
                tree.smart_add(["tree/link/file"]), (["dir", "dir/file"], {})
            )
        else:
            self.assertEqual(tree.smart_add(["tree/link/file"]), (["dir/file"], {}))

        # should add the actual parent directory, not the apparent parent
        # (which is actually a symlink)
        self.assertTrue(tree.is_versioned("dir/file"))
        self.assertTrue(tree.is_versioned("dir"))
        self.assertFalse(tree.is_versioned("link"))
        self.assertFalse(tree.is_versioned("link/file"))


class TestKindChanges(TestCaseWithWorkingTree):
    def setUp(self):
        super().setUp()
        self.requireFeature(features.SymlinkFeature(self.test_dir))

    def test_symlink_changes_to_dir(self):
        # <https://bugs.launchpad.net/bzr/+bug/192859>:
        # we had some past problems with the workingtree remembering for too
        # long what kind of object was at a particular name; we really
        # shouldn't do that.  Operating on the dirstate through passing
        # inventory deltas rather than mutating the inventory largely avoids
        # that.
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/a@", "target")])
        tree.smart_add(["tree/a"])
        tree.commit("add symlink")
        os.unlink("tree/a")
        self.build_tree_contents(
            [
                ("tree/a/",),
                ("tree/a/f", b"content"),
            ]
        )
        tree.smart_add(["tree/a/f"])
        tree.commit("change to dir")
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], list(tree.iter_changes(tree.basis_tree())))
        self.assertEqual(["a", "a/f"], sorted(info[0] for info in tree.list_files()))

    def test_dir_changes_to_symlink(self):
        # <https://bugs.launchpad.net/bzr/+bug/192859>:
        # we had some past problems with the workingtree remembering for too
        # long what kind of object was at a particular name; we really
        # shouldn't do that.  Operating on the dirstate through passing
        # inventory deltas rather than mutating the inventory largely avoids
        # that.
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents(
            [
                ("tree/a/",),
                ("tree/a/file", b"content"),
            ]
        )
        tree.smart_add(["tree/a"])
        tree.commit("add dir")
        osutils.rmtree("tree/a")
        self.build_tree_contents(
            [
                ("tree/a@", "target"),
            ]
        )
        tree.commit("change to symlink")


class TestOpenTree(TestCaseWithWorkingTree):
    def setUp(self):
        super().setUp()
        self.requireFeature(features.SymlinkFeature(self.test_dir))

    def test_open_containing_through_symlink(self):
        self.make_test_tree()
        self.check_open_containing("link/content", "tree", "content")
        self.check_open_containing("link/sublink", "tree", "sublink")
        # this next one is a bit debatable, but arguably it's better that
        # open_containing is only concerned with opening the tree
        # and then you can deal with symlinks along the way if you want
        self.check_open_containing(
            "link/sublink/subcontent", "tree", "sublink/subcontent"
        )

    def check_open_containing(self, to_open, expected_tree_name, expected_relpath):
        wt, relpath = workingtree.WorkingTree.open_containing(to_open)
        self.assertEqual(relpath, expected_relpath)
        self.assertEndsWith(wt.basedir, expected_tree_name)

    def test_tree_files(self):
        # not strictly a WorkingTree method, but it should be
        # probably the root cause for
        # <https://bugs.launchpad.net/bzr/+bug/128562>
        self.make_test_tree()
        self.check_tree_files(["tree/outerlink"], "tree", ["outerlink"])
        self.check_tree_files(["link/outerlink"], "tree", ["outerlink"])
        self.check_tree_files(
            ["link/sublink/subcontent"], "tree", ["subdir/subcontent"]
        )

    def check_tree_files(self, to_open, expected_tree, expect_paths):
        tree, relpaths = workingtree.WorkingTree.open_containing_paths(to_open)
        self.assertEndsWith(tree.basedir, expected_tree)
        self.assertEqual(expect_paths, relpaths)

    def make_test_tree(self):
        self.make_branch_and_tree("tree")
        self.build_tree_contents(
            [
                ("link@", "tree"),
                ("tree/outerlink@", "/not/there"),
                ("tree/content", b"hello"),
                ("tree/sublink@", "subdir"),
                ("tree/subdir/",),
                ("tree/subdir/subcontent", b"subcontent stuff"),
            ]
        )



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.annotate_iter'."""

from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestAnnotateIter(TestCaseWithWorkingTree):
    def make_single_rev_tree(self):
        builder = self.make_branch_builder("branch")
        revid = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", None)),
                ("add", ("file", None, "file", b"initial content\n")),
            ],
        )
        b = builder.get_branch()
        tree = b.create_checkout("tree", lightweight=True)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        return tree, revid

    def test_annotate_same_as_parent(self):
        tree, revid = self.make_single_rev_tree()
        annotations = tree.annotate_iter("file")
        self.assertEqual([(revid, b"initial content\n")], annotations)

    def test_annotate_mod_from_parent(self):
        tree, revid = self.make_single_rev_tree()
        self.build_tree_contents([("tree/file", b"initial content\nnew content\n")])
        annotations = tree.annotate_iter("file")
        self.assertEqual(
            [
                (revid, b"initial content\n"),
                (b"current:", b"new content\n"),
            ],
            annotations,
        )

    def test_annotate_merge_parents(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        revid1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", None)),
                ("add", ("file", None, "file", b"initial content\n")),
            ],
        )
        revid2 = builder.build_snapshot(
            [revid1],
            [
                ("modify", ("file", b"initial content\ncontent in 2\n")),
            ],
        )
        revid3 = builder.build_snapshot(
            [revid1],
            [
                ("modify", ("file", b"initial content\ncontent in 3\n")),
            ],
        )
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout("tree", revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents(
            [
                (
                    "tree/file",
                    b"initial content\ncontent in 2\ncontent in 3\nnew content\n",
                )
            ]
        )
        annotations = tree.annotate_iter("file")
        self.assertEqual(
            [
                (revid1, b"initial content\n"),
                (revid2, b"content in 2\n"),
                (revid3, b"content in 3\n"),
                (b"current:", b"new content\n"),
            ],
            annotations,
        )

    def test_annotate_merge_parent_no_file(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        revid1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", None)),
            ],
        )
        revid2 = builder.build_snapshot(
            [revid1],
            [
                ("add", ("file", None, "file", b"initial content\n")),
            ],
        )
        revid3 = builder.build_snapshot([revid1], [])
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout("tree", revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([("tree/file", b"initial content\nnew content\n")])
        annotations = tree.annotate_iter("file")
        self.assertEqual(
            [
                (revid2, b"initial content\n"),
                (b"current:", b"new content\n"),
            ],
            annotations,
        )

    def test_annotate_merge_parent_was_directory(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        revid1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", None)),
            ],
        )
        revid2 = builder.build_snapshot(
            [revid1],
            [
                ("add", ("file", None, "file", b"initial content\n")),
            ],
        )
        revid3 = builder.build_snapshot(
            [revid1],
            [
                ("add", ("a_dir", None, "directory", None)),
            ],
        )
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout("tree", revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([("tree/file", b"initial content\nnew content\n")])
        annotations = tree.annotate_iter("file")
        self.assertEqual(
            [
                (revid2, b"initial content\n"),
                (b"current:", b"new content\n"),
            ],
            annotations,
        )

    def test_annotate_same_as_merge_parent(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        revid1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", None)),
                ("add", ("file", None, "file", b"initial content\n")),
            ],
        )
        revid2 = builder.build_snapshot([revid1], [])
        revid3 = builder.build_snapshot(
            [revid1],
            [
                ("modify", ("file", b"initial content\ncontent in 3\n")),
            ],
        )
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout("tree", revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid3])
        self.build_tree_contents([("tree/file", b"initial content\ncontent in 3\n")])
        annotations = tree.annotate_iter("file")
        self.assertEqual(
            [
                (revid1, b"initial content\n"),
                (revid3, b"content in 3\n"),
            ],
            annotations,
        )

    def test_annotate_same_as_merge_parent_supersedes(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        revid1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", None)),
                ("add", ("file", None, "file", b"initial content\n")),
            ],
        )
        revid2 = builder.build_snapshot(
            [revid1],
            [
                ("modify", ("file", b"initial content\nnew content\n")),
            ],
        )
        revid3 = builder.build_snapshot(
            [revid2],
            [
                ("modify", ("file", b"initial content\ncontent in 3\n")),
            ],
        )
        revid4 = builder.build_snapshot(
            [revid3],
            [
                ("modify", ("file", b"initial content\nnew content\n")),
            ],
        )
        # In this case, the content locally is the same as content in basis
        # tree, but the merge revision states that *it* should win
        builder.finish_series()
        b = builder.get_branch()
        tree = b.create_checkout("tree", revision_id=revid2, lightweight=True)
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.set_parent_ids([revid2, revid4])
        annotations = tree.annotate_iter("file")
        self.assertEqual(
            [
                (revid1, b"initial content\n"),
                (revid4, b"new content\n"),
            ],
            annotations,
        )



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Symbol versioning tests."""

import sys
import warnings

from breezy import symbol_versioning
from breezy.symbol_versioning import (
    deprecated_function,
    deprecated_in,
    deprecated_method,
)
from breezy.tests import TestCase


@deprecated_function(deprecated_in((0, 7, 0)))
def sample_deprecated_function():
    """Deprecated function docstring."""
    return 1


a_deprecated_list = symbol_versioning.deprecated_list(
    deprecated_in((0, 9, 0)), "a_deprecated_list", ["one"], extra="Don't use me"
)


a_deprecated_dict = symbol_versioning.DeprecatedDict(
    deprecated_in((0, 14, 0)),
    "a_deprecated_dict",
    {"a": 42},
    advice="Pull the other one!",
)


class TestDeprecationWarnings(TestCase):
    def capture_warning(self, message, category, stacklevel=None):
        self._warnings.append((message, category, stacklevel))

    def setUp(self):
        super().setUp()
        self._warnings = []

    @deprecated_method(deprecated_in((0, 7, 0)))
    def deprecated_method(self):
        """Deprecated method docstring.

        This might explain stuff.
        """
        return 1

    @staticmethod
    @deprecated_function(deprecated_in((0, 7, 0)))
    def deprecated_static():
        """Deprecated static."""
        return 1

    def test_deprecated_static(self):
        expected_warning = (
            "breezy.tests.test_symbol_versioning.TestDeprecationWarnings."
            "deprecated_static "
            "was deprecated in version 0.7.0.",
            DeprecationWarning,
            2,
        )
        expected_docstring = (
            "Deprecated static.\n\nThis function was deprecated in version 0.7.0.\n"
        )
        self.check_deprecated_callable(
            expected_warning,
            expected_docstring,
            "deprecated_static",
            "breezy.tests.test_symbol_versioning",
            self.deprecated_static,
        )

    def test_deprecated_method(self):
        expected_warning = (
            "breezy.tests.test_symbol_versioning."
            "TestDeprecationWarnings.deprecated_method "
            "was deprecated in version 0.7.0.",
            DeprecationWarning,
            2,
        )
        if sys.version_info >= (3, 13):
            expected_docstring = (
                "Deprecated method docstring.\n"
                "\n"
                "This might explain stuff.\n"
                "\n"
                "This method was deprecated in version 0.7.0.\n"
                ""
            )
        else:
            expected_docstring = (
                "Deprecated method docstring.\n"
                "\n"
                "        This might explain stuff.\n"
                "        \n"
                "        This method was deprecated in version 0.7.0.\n"
                "        "
            )
        self.check_deprecated_callable(
            expected_warning,
            expected_docstring,
            "deprecated_method",
            "breezy.tests.test_symbol_versioning",
            self.deprecated_method,
        )

    def test_deprecated_function(self):
        expected_warning = (
            "breezy.tests.test_symbol_versioning.sample_deprecated_function "
            "was deprecated in version 0.7.0.",
            DeprecationWarning,
            2,
        )
        expected_docstring = (
            "Deprecated function docstring.\n"
            "\n"
            "This function was deprecated in version 0.7.0.\n"
        )
        self.check_deprecated_callable(
            expected_warning,
            expected_docstring,
            "sample_deprecated_function",
            "breezy.tests.test_symbol_versioning",
            sample_deprecated_function,
        )

    def test_deprecated_list(self):
        expected_warning = (
            "Modifying a_deprecated_list was deprecated in version 0.9.0. Don't use me",
            DeprecationWarning,
            3,
        )
        old_warning_method = symbol_versioning.warn
        try:
            symbol_versioning.set_warning_method(self.capture_warning)
            self.assertEqual(["one"], a_deprecated_list)
            self.assertEqual([], self._warnings)

            a_deprecated_list.append("foo")
            self.assertEqual([expected_warning], self._warnings)
            self.assertEqual(["one", "foo"], a_deprecated_list)

            a_deprecated_list.extend(["bar", "baz"])
            self.assertEqual([expected_warning] * 2, self._warnings)
            self.assertEqual(["one", "foo", "bar", "baz"], a_deprecated_list)

            a_deprecated_list.insert(1, "xxx")
            self.assertEqual([expected_warning] * 3, self._warnings)
            self.assertEqual(["one", "xxx", "foo", "bar", "baz"], a_deprecated_list)

            a_deprecated_list.remove("foo")
            self.assertEqual([expected_warning] * 4, self._warnings)
            self.assertEqual(["one", "xxx", "bar", "baz"], a_deprecated_list)

            val = a_deprecated_list.pop()
            self.assertEqual([expected_warning] * 5, self._warnings)
            self.assertEqual("baz", val)
            self.assertEqual(["one", "xxx", "bar"], a_deprecated_list)

            val = a_deprecated_list.pop(1)
            self.assertEqual([expected_warning] * 6, self._warnings)
            self.assertEqual("xxx", val)
            self.assertEqual(["one", "bar"], a_deprecated_list)
        finally:
            symbol_versioning.set_warning_method(old_warning_method)

    def test_deprecated_dict(self):
        expected_warning = (
            "access to a_deprecated_dict was deprecated in version 0.14.0."
            " Pull the other one!",
            DeprecationWarning,
            2,
        )
        old_warning_method = symbol_versioning.warn
        try:
            symbol_versioning.set_warning_method(self.capture_warning)
            self.assertEqual(len(a_deprecated_dict), 1)
            self.assertEqual([expected_warning], self._warnings)

            a_deprecated_dict["b"] = 42
            self.assertEqual(a_deprecated_dict["b"], 42)
            self.assertTrue("b" in a_deprecated_dict)
            del a_deprecated_dict["b"]
            self.assertFalse("b" in a_deprecated_dict)
            self.assertEqual([expected_warning] * 6, self._warnings)
        finally:
            symbol_versioning.set_warning_method(old_warning_method)

    def check_deprecated_callable(
        self,
        expected_warning,
        expected_docstring,
        expected_name,
        expected_module,
        deprecated_callable,
    ):
        if __doc__ is None:
            # With -OO the docstring should just be the deprecated version
            expected_docstring = expected_docstring.split("\n")[-2].lstrip()
        old_warning_method = symbol_versioning.warn
        try:
            symbol_versioning.set_warning_method(self.capture_warning)
            self.assertEqual(1, deprecated_callable())
            self.assertEqual([expected_warning], self._warnings)
            deprecated_callable()
            self.assertEqual([expected_warning, expected_warning], self._warnings)
            self.assertEqualDiff(expected_docstring, deprecated_callable.__doc__)
            self.assertEqualDiff(expected_name, deprecated_callable.__name__)
            self.assertEqualDiff(expected_module, deprecated_callable.__module__)
            self.assertTrue(deprecated_callable.is_deprecated)
        finally:
            symbol_versioning.set_warning_method(old_warning_method)

    def test_deprecated_passed(self):
        self.assertEqual(True, symbol_versioning.deprecated_passed(None))
        self.assertEqual(True, symbol_versioning.deprecated_passed(True))
        self.assertEqual(True, symbol_versioning.deprecated_passed(False))
        self.assertEqual(
            False,
            symbol_versioning.deprecated_passed(symbol_versioning.DEPRECATED_PARAMETER),
        )

    def test_deprecation_string(self):
        """We can get a deprecation string for a method or function."""
        err_message = symbol_versioning.deprecation_string(
            self.test_deprecation_string, deprecated_in((0, 11, 0))
        )
        self.assertEqual(
            err_message,
            "breezy.tests.test_symbol_versioning.TestDeprecationWarnings."
            "test_deprecation_string was deprecated in "
            "version 0.11.0.",
        )

        self.assertEqual(
            "breezy.symbol_versioning.deprecated_function was "
            "deprecated in version 0.11.0.",
            symbol_versioning.deprecation_string(
                symbol_versioning.deprecated_function, deprecated_in((0, 11, 0))
            ),
        )


class TestSuppressAndActivate(TestCase):
    def setUp(self):
        super().setUp()
        existing_filters = list(warnings.filters)

        def restore():
            warnings.filters[:] = existing_filters

        self.addCleanup(restore)
        # Clean out the filters so we have a clean slate.
        warnings.resetwarnings()

    def assertFirstWarning(self, action, category):
        """Test the first warning in the filters is correct."""
        first = warnings.filters[0]
        self.assertEqual((action, category), (first[0], first[2]))

    def test_suppress_deprecation_warnings(self):
        """suppress_deprecation_warnings sets DeprecationWarning to ignored."""
        symbol_versioning.suppress_deprecation_warnings()
        self.assertFirstWarning("ignore", DeprecationWarning)

    def test_set_restore_filters(self):
        original_filters = warnings.filters[:]
        symbol_versioning.suppress_deprecation_warnings()()
        self.assertEqual(original_filters, warnings.filters)

    def test_suppress_deprecation_with_warning_filter(self):
        """Don't suppress if we already have a filter."""
        warnings.filterwarnings("error", category=Warning)
        self.assertFirstWarning("error", Warning)
        self.assertEqual(1, len(warnings.filters))
        symbol_versioning.suppress_deprecation_warnings(override=False)
        self.assertFirstWarning("error", Warning)
        self.assertEqual(1, len(warnings.filters))

    def test_suppress_deprecation_with_filter(self):
        """Don't suppress if we already have a filter."""
        warnings.filterwarnings("error", category=DeprecationWarning)
        self.assertFirstWarning("error", DeprecationWarning)
        self.assertEqual(1, len(warnings.filters))
        symbol_versioning.suppress_deprecation_warnings(override=False)
        self.assertFirstWarning("error", DeprecationWarning)
        self.assertEqual(1, len(warnings.filters))
        symbol_versioning.suppress_deprecation_warnings(override=True)
        self.assertFirstWarning("ignore", DeprecationWarning)
        self.assertEqual(2, len(warnings.filters))

    def test_activate_deprecation_no_error(self):
        # First nuke the filters, so we know it is clean
        symbol_versioning.activate_deprecation_warnings()
        self.assertFirstWarning("default", DeprecationWarning)

    def test_activate_deprecation_with_error(self):
        # First nuke the filters, so we know it is clean
        # Add a warning == error rule
        warnings.filterwarnings("error", category=Warning)
        self.assertFirstWarning("error", Warning)
        self.assertEqual(1, len(warnings.filters))
        symbol_versioning.activate_deprecation_warnings(override=False)
        # There should not be a new warning
        self.assertFirstWarning("error", Warning)
        self.assertEqual(1, len(warnings.filters))

    def test_activate_deprecation_with_DW_error(self):
        # First nuke the filters, so we know it is clean
        # Add a warning == error rule
        warnings.filterwarnings("error", category=DeprecationWarning)
        self.assertFirstWarning("error", DeprecationWarning)
        self.assertEqual(1, len(warnings.filters))
        symbol_versioning.activate_deprecation_warnings(override=False)
        # There should not be a new warning
        self.assertFirstWarning("error", DeprecationWarning)
        self.assertEqual(1, len(warnings.filters))
        symbol_versioning.activate_deprecation_warnings(override=True)
        self.assertFirstWarning("default", DeprecationWarning)
        self.assertEqual(2, len(warnings.filters))



# Copyright (C) 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import errors, tests
from breezy.tests.per_tree import TestCaseWithTree


class Path2IdTests(TestCaseWithTree):
    def setUp(self):
        super().setUp()
        work_a = self.make_branch_and_tree("wta")
        if not work_a.supports_setting_file_ids():
            self.skipTest("working tree does not support setting file ids")
        self.build_tree(["wta/bla", "wta/dir/", "wta/dir/file"])
        work_a.add(["bla", "dir", "dir/file"], ids=[b"bla-id", b"dir-id", b"file-id"])
        work_a.commit("add files")
        self.tree_a = self.workingtree_to_test_tree(work_a)

    def test_path2id(self):
        self.assertEqual(b"bla-id", self.tree_a.path2id("bla"))
        self.assertEqual(b"dir-id", self.tree_a.path2id("dir"))
        self.assertIs(None, self.tree_a.path2id("idontexist"))

    def test_path2id_list(self):
        self.assertEqual(b"bla-id", self.tree_a.path2id(["bla"]))
        self.assertEqual(b"dir-id", self.tree_a.path2id(["dir"]))
        self.assertEqual(b"file-id", self.tree_a.path2id(["dir", "file"]))
        self.assertEqual(self.tree_a.path2id(""), self.tree_a.path2id([]))
        self.assertIs(None, self.tree_a.path2id(["idontexist"]))
        self.assertIs(None, self.tree_a.path2id(["dir", "idontexist"]))

    def test_id2path(self):
        self.addCleanup(self.tree_a.lock_read().unlock)
        self.assertEqual("bla", self.tree_a.id2path(b"bla-id"))
        self.assertEqual("dir", self.tree_a.id2path(b"dir-id"))
        self.assertEqual("dir/file", self.tree_a.id2path(b"file-id"))
        self.assertRaises(errors.NoSuchId, self.tree_a.id2path, b"nonexistant")

    def skip_if_no_reference(self, tree):
        if not getattr(tree, "supports_tree_reference", lambda: False)():
            raise tests.TestNotApplicable("Tree references not supported")

    def create_nested(self):
        work_tree = self.make_branch_and_tree("wt")
        with work_tree.lock_write():
            self.skip_if_no_reference(work_tree)
            subtree = self.make_branch_and_tree("wt/subtree")
            self.build_tree(["wt/subtree/a"])
            subtree.add(["a"])
            subtree.commit("foo")
            work_tree.add_reference(subtree)
        tree = self._convert_tree(work_tree)
        self.skip_if_no_reference(tree)
        return tree, subtree

    def test_path2id_nested_tree(self):
        tree, subtree = self.create_nested()
        self.assertIsNot(None, tree.path2id("subtree"))
        self.assertIsNot(None, tree.path2id("subtree/a"))
        self.assertEqual("subtree", tree.id2path(tree.path2id("subtree")))
        self.assertEqual("subtree/a", tree.id2path(tree.path2id("subtree/a")))
        self.assertIsNot(
            "subtree/a", tree.id2path(tree.path2id("subtree/a"), recurse="down")
        )
        self.assertRaises(
            errors.NoSuchId, tree.id2path, tree.path2id("subtree/a"), recurse="none"
        )


class Path2IdsTests(TestCaseWithTree):
    def test_paths2ids_recursive(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/dir/", "tree/dir/file"])
        work_tree.add(["dir", "dir/file"])
        if not work_tree.supports_setting_file_ids():
            raise tests.TestNotApplicable("test not applicable on non-inventory tests")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            {tree.path2id("dir"), tree.path2id("dir/file")}, tree.paths2ids(["dir"])
        )

    def test_paths2ids_forget_old(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file"])
        work_tree.add("file")
        work_tree.commit("commit old state")
        work_tree.remove("file")
        if not work_tree.supports_setting_file_ids():
            raise tests.TestNotApplicable("test not applicable on non-inventory tests")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(set(), tree.paths2ids(["file"], require_versioned=False))
